<?php
/**
 * The Horde_Text_Rst_Parser_Node class represents a parse node of a reStructuredText
 * document.
 *
 * $Horde: framework/Text_Rst/lib/Horde/Text/Rst/Parser/Node.php,v 1.2.2.2 2008/01/02 11:30:57 jan Exp $
 *
 * Copyright 2003-2008 The Horde Project (http://www.horde.org/)
 *
 * See the enclosed file COPYING for license information (LGPL). If you
 * did not receive this file, see http://opensource.org/licenses/lgpl-license.php.
 *
 * @author   Jason M. Felice <jason.m.felice@gmail.com>
 * @author   Chuck Hagenbuch <chuck@horde.org>
 * @category Horde
 * @package  Horde_Text_Rst
 */
class Horde_Text_Rst_Parser_Node implements IteratorAggregate
{
    /**
     * This node's type
     *
     * @var string
     */
    protected $_type = 'document';

    /**
     * An array of this parse node's children.
     *
     * @var array
     */
    public $children = array();

    /**
     * Constructor.
     *
     * @param string $type  This is the node type.  The default is 'document'.
     */
    public function __construct($type = 'document')
    {
        $this->_type = $type;
    }

    /**
     * @return string The node's type
     */
    public function getType()
    {
        return $this->_type;
    }

    /**
     * Appends a child parse node to this parse node.
     *
     * @param string|Horde_Text_Rst $child  This is the string or object child
     *                                 to append to this parse node.
     */
    public function appendChild($child)
    {
        $n = count($this->children);
        if (is_string($child) && $n > 0 && is_string($this->children[$n - 1])) {
            $this->children[$n - 1] .= $child;
        } else {
            $this->children[] = $child;
        }

        if ($child instanceof Horde_Text_Rst_Parser_Node) {
            $child->parent = $this;
        }
    }

    /**
     * Return an iterator over this node's properties
     *
     * @return ArrayIterator
     */
    public function getIterator()
    {
        $vars = get_object_vars($this);
        unset($vars['_type'], $vars['children'], $vars['parent']);
        ksort($vars);

        return new ArrayIterator($vars);
    }

    /**
     * Dumps this parse node and its children.
     * This method is for debugging purposes.
     *
     * @param integer $level  This is the indent level of this parse node.
     *
     * @return string The node dump.
     */
    public function dump($level = 0)
    {
        $dump = str_repeat('    ', $level)
            . '<' . $this->_type;
        foreach ($this as $name => $value) {
            $dump .= ' ' . $name . '="'
                . preg_replace('/["\\\\]/', '\\$1', $value) . '"';
        }
        $dump .= ">\n";
        foreach ($this->children as $child) {
            if (is_string($child)) {
                $dump .= str_repeat('    ', $level + 1)
                    . str_replace("\n", "\n" . str_repeat('    ', $level + 1),
                                  preg_replace('/["\\\\]/', '\\$1', $child)) . "\n";
            } else {
                $dump .= $child->dump($level + 1);
            }
        }

        return $dump;
    }

}
